import asyncio
from asyncio import CancelledError, Future, InvalidStateError
from binascii import hexlify
from contextlib import suppress
from dataclasses import dataclass
import errno
from functools import reduce
from io import BytesIO
from ipaddress import ip_address
import logging
from operator import xor
import socket
import struct
from typing import Dict, Optional, Union

from construct import (
    Adapter,
    Array,
    BitStruct,
    Bytes,
    Checksum,
    ChecksumError,
    Const,
    Container,
    Enum,
    EnumIntegerString,
    FixedSized,
    Flag,
    FlagsEnum,
    GreedyBytes,
    GreedyString,
    IfThenElse,
    Int8sb,
    Int8ub,
    Int16sl,
    Int16ub,
    Int16ul,
    NullTerminated,
    Pointer,
    Prefixed,
    RawCopy,
    Select,
    StringEncoded,
    Struct,
    Subconstruct,
    Switch,
    Union as UnionConstruct,
    this,
)
from tenacity import retry, retry_if_exception_type, stop_after_attempt

from nibe.coil import Coil, CoilData
from nibe.connection import DEFAULT_TIMEOUT, READ_PRODUCT_INFO_TIMEOUT, Connection
from nibe.connection.encoders import CoilDataEncoderNibeGw
from nibe.connection.mixins import ConnectionStatus, ConnectionStatusMixin
from nibe.event_server import EventServer
from nibe.exceptions import (
    AddressInUseException,
    CoilNotFoundException,
    CoilWriteSendException,
    DecodeException,
    NibeException,
    ProductInfoReadTimeoutException,
    ReadException,
    ReadIOException,
    ReadSendException,
    ReadTimeoutException,
    WriteException,
    WriteIOException,
    WriteTimeoutException,
)
from nibe.heatpump import HeatPump, ProductInfo

from . import verify_connectivity_read_write_alarm

logger = logging.getLogger("nibe").getChild(__name__)


@dataclass
class CoilAction:
    coil: Coil
    future: Future


class NibeGW(asyncio.DatagramProtocol, Connection, EventServer, ConnectionStatusMixin):
    """NibeGW connection."""

    PRODUCT_INFO_EVENT = "product_info"
    _futures: Dict[str, Future]
    _registered_reads: Dict[str, CoilAction]

    def __init__(
        self,
        heatpump: HeatPump,
        remote_ip: Optional[str] = None,
        remote_read_port: int = 9999,
        remote_write_port: int = 10000,
        listening_ip: str = "0.0.0.0",
        listening_port: int = 9999,
        read_retries: int = 3,
        write_retries: int = 3,
    ) -> None:
        super().__init__()

        self._heatpump = heatpump
        self._listening_ip = listening_ip
        self._listening_port = listening_port

        self._remote_ip = remote_ip
        self._remote_read_port = remote_read_port
        self._remote_write_port = remote_write_port

        self._transport = None

        self._send_lock = asyncio.Lock()
        self._futures = {}
        self._registered_reads = {}

        self.coil_encoder = CoilDataEncoderNibeGw(heatpump.word_swap)

        self.read_coil = retry(
            retry=retry_if_exception_type(ReadIOException),
            stop=stop_after_attempt(read_retries),
            reraise=True,
        )(self.read_coil)

        self.write_coil = retry(
            retry=retry_if_exception_type(WriteIOException),
            stop=stop_after_attempt(write_retries),
            reraise=True,
        )(self.write_coil)

    async def start(self):
        logger.info(f"Starting UDP server on port {self._listening_port}")

        self.status = ConnectionStatus.INITIALIZING

        family, type, proto, _, sockaddr = socket.getaddrinfo(
            self._listening_ip,
            self._listening_port,
            type=socket.SOCK_DGRAM,
            proto=socket.IPPROTO_UDP,
            family=socket.AddressFamily.AF_INET,
        )[0]

        sock = socket.socket(family, type, proto)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        try:
            if ip_address(sockaddr[0]).is_multicast:
                group_bin = socket.inet_pton(family, sockaddr[0])
                if family == socket.AF_INET:  # IPv4
                    sock.bind(("", sockaddr[1]))
                    mreq = group_bin + struct.pack("=I", socket.INADDR_ANY)
                    sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
                else:
                    sock.bind(("", sockaddr[1]))
                    mreq = group_bin + struct.pack("@I", 0)
                    sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_JOIN_GROUP, mreq)
            elif self._listening_ip:
                sock.bind(sockaddr)
            else:
                sock.bind(("", sockaddr[1]))
        except OSError as exception:
            if exception.errno == errno.EADDRINUSE:
                raise AddressInUseException(f"Address in use {sockaddr}")
            raise

        await asyncio.get_event_loop().create_datagram_endpoint(lambda: self, sock=sock)

        if self._heatpump.word_swap is None:
            await self.detect_word_swap()

    def connection_made(self, transport):
        """Callback when connection is made."""
        self.status = ConnectionStatus.LISTENING
        self._transport = transport

    def datagram_received(self, data: bytes, addr):
        """Callback when data is received."""
        logger.debug(f"Received {hexlify(data).decode('utf-8')} from {addr}")
        try:
            msg = Response.parse(data)

            if not self._remote_ip:
                logger.debug("Pump discovered at %s", addr)
                self._remote_ip = addr[0]

            self.status = ConnectionStatus.CONNECTED

            logger.debug(msg.fields.value)
            cmd = msg.fields.value.cmd
            if cmd == "MODBUS_DATA_MSG":
                for row in msg.fields.value.data:
                    self._on_raw_coil_value(row.coil_address, row.value)
            elif cmd == "MODBUS_READ_RESP":
                row = msg.fields.value.data
                self._on_raw_coil_value(row.coil_address, row.value)
            elif cmd == "MODBUS_WRITE_RESP":
                with suppress(InvalidStateError, CancelledError, KeyError):
                    self._futures["write"].set_result(msg.fields.value.data.result)
            elif cmd == "RMU_DATA_MSG":
                self._on_rmu_data(msg.fields.value)
            elif cmd == "PRODUCT_INFO_MSG":
                data = msg.fields.value.data
                product_info = ProductInfo(data["model"], data["version"])
                with suppress(InvalidStateError, CancelledError, KeyError):
                    self._futures["product_info"].set_result(product_info)
                self.notify_event_listeners(
                    self.PRODUCT_INFO_EVENT, product_info=product_info
                )
            elif not isinstance(cmd, EnumIntegerString):
                logger.debug(f"Unknown command {cmd}")
        except ChecksumError:
            logger.warning(
                f"Ignoring packet from {addr} due to checksum error: {hexlify(data).decode('utf-8')}"
            )
        except NibeException as e:
            logger.error(f"Failed handling packet from {addr}: {e}")
        except Exception as e:
            logger.exception(
                f"Unexpected exception during parsing packet data '{hexlify(data).decode('utf-8')}' from {addr}",
                e,
            )

    async def read_product_info(
        self, timeout: float = READ_PRODUCT_INFO_TIMEOUT
    ) -> ProductInfo:
        self._futures["product_info"] = asyncio.get_event_loop().create_future()
        try:
            return await asyncio.wait_for(self._futures["product_info"], timeout)
        except asyncio.TimeoutError:
            raise ProductInfoReadTimeoutException("Timeout waiting for product message")
        finally:
            del self._futures["product_info"]

    async def detect_word_swap(self, timeout: float = DEFAULT_TIMEOUT) -> None:
        """Read word swap setting."""
        try:
            coil = self._heatpump.get_coil_by_address(48852)
            assert coil.is_boolean, "Coil is not boolean"
            coil_data = await self.read_coil(coil, timeout)
            self.coil_encoder.word_swap = coil_data.value == "ON"
            self._heatpump.word_swap = self.coil_encoder.word_swap
            logger.info(f"Word swap setting detected: {coil_data.value}")
        except Exception as e:
            logger.warning(f"Failed to detect word swap setting: {e}")

    async def read_coil(self, coil: Coil, timeout: float = DEFAULT_TIMEOUT) -> CoilData:
        async with self._send_lock:
            assert self._transport, "Transport is closed"
            data = Request.build(
                dict(
                    fields=dict(
                        value=dict(
                            cmd="MODBUS_READ_REQ", data=dict(coil_address=coil.address)
                        )
                    )
                )
            )

            future = self._register_coil_read_request(coil)

            logger.debug(
                f"Sending {hexlify(data).decode('utf-8')} (read request) to {self._remote_ip}:{self._remote_write_port}"
            )

            try:
                self._transport.sendto(data, (self._remote_ip, self._remote_read_port))
            except socket.gaierror:
                raise ReadSendException(f"Unable to lookup hostname: {self._remote_ip}")

            logger.debug(f"Waiting for read response for {coil.name}")

            try:
                return await asyncio.wait_for(future, timeout)
            except asyncio.TimeoutError:
                raise ReadTimeoutException(
                    f"Timeout waiting for read response for {coil.name}"
                )
            except DecodeException as e:
                raise ReadException(
                    f"Failed decoding response for {coil.name}: {e}"
                ) from e

    def _register_coil_read_request(self, coil: Coil) -> Future:
        read = self._registered_reads.get(str(coil.address))
        if read is not None and not read.future.done():
            return read.future

        future = asyncio.get_event_loop().create_future()
        self._registered_reads[str(coil.address)] = CoilAction(coil, future)
        return future

    def _on_coil_read_success(self, coil_data):
        logger.info(coil_data)

        read = self._registered_reads.get(str(coil_data.coil.address))
        if read is not None and not read.future.done():
            read.future.set_result(coil_data)

        self._heatpump.notify_coil_update(coil_data)

    def _on_coil_read_error(
        self, coil_address, value: Union[bytes, float, int, str], exception: Exception
    ):
        if coil_address == 65535:  # 0xffff
            return

        read = self._registered_reads.get(str(coil_address))
        if read is not None and not read.future.done():
            read.future.set_exception(exception)

        if isinstance(exception, CoilNotFoundException):
            logger.warning(f"Ignoring coil {coil_address} - coil definition not found")
        elif isinstance(exception, DecodeException):
            str_value = (
                hexlify(value).decode("utf-8") if isinstance(value, bytes) else value
            )
            logger.warning(
                f"Ignoring coil {coil_address} value {str_value} - failed to decode"
            )
        elif isinstance(exception, NibeException):
            logger.error(f"Failed handling read for {coil_address}: {exception}")

    async def write_coil(
        self, coil_data: CoilData, timeout: float = DEFAULT_TIMEOUT
    ) -> None:
        coil = coil_data.coil
        assert coil.is_writable, f"{coil.name} is not writable"
        assert coil_data.value is not None, f"{coil.name} value must be set"
        async with self._send_lock:
            assert self._transport, "Transport is closed"
            data = Request.build(
                dict(
                    fields=dict(
                        value=dict(
                            cmd="MODBUS_WRITE_REQ",
                            data=dict(
                                coil_address=coil.address,
                                value=self.coil_encoder.encode(coil_data),
                            ),
                        )
                    )
                )
            )

            self._futures["write"] = asyncio.get_event_loop().create_future()

            logger.debug(
                f"Sending {hexlify(data).decode('utf-8')} (write request) to {self._remote_ip}:{self._remote_write_port}"
            )

            try:
                self._transport.sendto(data, (self._remote_ip, self._remote_write_port))
            except socket.gaierror:
                raise CoilWriteSendException(
                    f"Unable to lookup hostname: {self._remote_ip}"
                )

            try:
                await asyncio.wait_for(self._futures["write"], timeout)

                result = self._futures["write"].result()

                if not result:
                    raise WriteException(f"Heatpump denied writing {coil.name}")
                else:
                    logger.info(f"Write succeeded for {coil.name}")
            except asyncio.TimeoutError:
                raise WriteTimeoutException(
                    f"Timeout waiting for write feedback for {coil.name}"
                )
            finally:
                del self._futures["write"]

    def error_received(self, exc):
        """Handle errors from the transport"""
        logger.error(exc)

    @property
    def remote_ip(self) -> Optional[str]:
        """Get the remote IP address"""
        return self._remote_ip

    def _on_rmu_data(self, value: Container):
        data = value.data
        self._on_coil_value(40004, data.bt1_outdoor_temperature)
        self._on_coil_value(40013, data.bt7_hw_top)

        if data.flags.use_room_sensor_s1:
            self._on_coil_value(47398, data.setpoint_or_offset_s1)
        else:
            self._on_coil_value(47011, data.setpoint_or_offset_s1)

        if data.flags.use_room_sensor_s2:
            self._on_coil_value(47397, data.setpoint_or_offset_s2)
        else:
            self._on_coil_value(47010, data.setpoint_or_offset_s2)

        if data.flags.use_room_sensor_s3:
            self._on_coil_value(47396, data.setpoint_or_offset_s3)
        else:
            self._on_coil_value(47009, data.setpoint_or_offset_s3)

        if data.flags.use_room_sensor_s4:
            self._on_coil_value(47395, data.setpoint_or_offset_s4)
        else:
            self._on_coil_value(47008, data.setpoint_or_offset_s4)

        if data.operational_mode == "MANUAL":
            self._on_coil_value(47370, data.flags.allow_additive_heating)
            self._on_coil_value(47371, data.flags.allow_heating)
            self._on_coil_value(47372, data.flags.allow_cooling)

        self._on_coil_value(48132, data.temporary_lux)
        self._on_coil_value(45001, data.alarm)
        self._on_coil_value(47137, data.operational_mode)
        self._on_coil_value(47387, data.flags.hw_production)

        if coil_address := ADDRESS_TO_ROOM_TEMP_COIL.get(value.address):
            self._on_coil_value(coil_address, data.bt50_room_temp_sX)

    def _on_raw_coil_value(self, coil_address: int, raw_value: bytes) -> None:
        try:
            coil = self._heatpump.get_coil_by_address(coil_address)
            coil_data = self.coil_encoder.decode(coil, raw_value)
            self._on_coil_read_success(coil_data)
        except NibeException as e:
            self._on_coil_read_error(coil_address, raw_value, e)

    def _on_coil_value(self, coil_address: int, value: Union[float, int, str]) -> None:
        try:
            coil = self._heatpump.get_coil_by_address(coil_address)

            if isinstance(value, bool):
                value = 1 if value else 0

            if isinstance(value, EnumIntegerString):
                value = int(value)

            if coil.has_mappings and isinstance(value, int):
                value = coil.get_mapping_for(value)

            coil_data = CoilData(coil, value)
            logger.info(coil_data)
            self._on_coil_read_success(coil_data)
        except NibeException as e:
            self._on_coil_read_error(coil_address, value, e)

    async def verify_connectivity(self):
        await verify_connectivity_read_write_alarm(self, self._heatpump)

    async def stop(self):
        if self._transport:
            self._transport.close()
            self._transport = None
        await asyncio.sleep(0)
        self.status = ConnectionStatus.DISCONNECTED


def xor8(data: bytes) -> int:
    chksum = reduce(xor, data)
    if chksum == 0x5C:
        chksum = 0xC5
    return chksum


class Dedupe5C(Subconstruct):
    def __init__(self, subcon):
        super().__init__(subcon)
        self.name = subcon.name

    def _parse(self, stream, context, path):
        unescaped = stream.getvalue().replace(b"\x5c\x5c", b"\x5c")
        context.length = len(unescaped)
        with BytesIO(unescaped) as stream2:
            obj = self.subcon._parsereport(stream2, context, path)
        return obj

    def _build(self, obj, stream, context, path):
        escaped = stream.getvalue().replace(b"\x5c", b"\x5c\x5c")
        context.length = len(escaped)
        with BytesIO(escaped) as stream2:
            self.subcon._build(obj, stream2, context, path)
        return obj


ProductInfoData = Struct(
    "_unknown" / Bytes(1), "version" / Int16ub, "model" / GreedyString("ASCII")
)


class FixedPoint(Adapter):
    def __init__(self, subcon, scale, offset, ndigits=1) -> None:
        super().__init__(subcon)
        self._offset = offset
        self._scale = scale
        self._ndigits = ndigits

    def _decode(self, obj, context, path):
        return round(obj * self._scale + self._offset, self._ndigits)

    def _encode(self, obj, context, path):
        return (obj - self._offset) / self._scale


StringData = Struct(
    "unknown" / Int8ub,
    "id" / Int16ul,
    "string" / StringEncoded(NullTerminated(GreedyBytes), "ISO-8859-1"),
)

RmuData = Struct(
    "flags"
    / Pointer(
        15,
        BitStruct(
            "unknown_8000" / Flag,
            "unknown_4000" / Flag,
            "unknown_2000" / Flag,
            "unknown_1000" / Flag,
            "unknown_0800" / Flag,
            "allow_cooling" / Flag,
            "allow_heating" / Flag,
            "allow_additive_heating" / Flag,
            "use_room_sensor_s4" / Flag,
            "use_room_sensor_s3" / Flag,
            "use_room_sensor_s2" / Flag,
            "use_room_sensor_s1" / Flag,
            "unknown_0008" / Flag,
            "unknown_0004" / Flag,
            "unknown_0002" / Flag,
            "hw_production" / Flag,
        ),
    ),
    "bt1_outdoor_temperature" / FixedPoint(Int16sl, 0.1, -0.5),
    "bt7_hw_top" / FixedPoint(Int16sl, 0.1, -0.5),
    "setpoint_or_offset_s1"
    / IfThenElse(
        lambda this: this.flags.use_room_sensor_s1,
        FixedPoint(Int8ub, 0.1, 5.0),
        FixedPoint(Int8sb, 0.1, 0),
    ),
    "setpoint_or_offset_s2"
    / IfThenElse(
        lambda this: this.flags.use_room_sensor_s2,
        FixedPoint(Int8ub, 0.1, 5.0),
        FixedPoint(Int8sb, 0.1, 0),
    ),
    "setpoint_or_offset_s3"
    / IfThenElse(
        lambda this: this.flags.use_room_sensor_s3,
        FixedPoint(Int8ub, 0.1, 5.0),
        FixedPoint(Int8sb, 0.1, 0),
    ),
    "setpoint_or_offset_s4"
    / IfThenElse(
        lambda this: this.flags.use_room_sensor_s4,
        FixedPoint(Int8ub, 0.1, 5.0),
        FixedPoint(Int8sb, 0.1, 0),
    ),
    "bt50_room_temp_sX" / FixedPoint(Int16sl, 0.1, -0.5),
    "temporary_lux" / Int8ub,
    "hw_time_hour" / Int8ub,
    "hw_time_min" / Int8ub,
    "fan_mode" / Int8ub,
    "operational_mode" / Int8ub,
    "_flags" / Bytes(2),
    "clock_time_hour" / Int8ub,
    "clock_time_min" / Int8ub,
    "alarm" / Int8ub,
    "unknown4" / Bytes(1),
    "fan_time_hour" / Int8ub,
    "fan_time_min" / Int8ub,
    "unknown5" / GreedyBytes,
)

Data = Dedupe5C(
    Switch(
        this.cmd,
        {
            "MODBUS_READ_RESP": Struct("coil_address" / Int16ul, "value" / Bytes(4)),
            "MODBUS_DATA_MSG": Array(
                lambda this: this.length // 4,
                Struct("coil_address" / Int16ul, "value" / Bytes(2)),
            ),
            "MODBUS_WRITE_RESP": Struct("result" / Flag),
            "MODBUS_ADDRESS_MSG": Struct("address" / Int8ub),
            "PRODUCT_INFO_MSG": ProductInfoData,
            "RMU_DATA_MSG": RmuData,
            "STRING_MSG": StringData,
        },
        default=Bytes(this.length),
    )
)


Command = Enum(
    Int8ub,
    RMU_WRITE_REQ=0x60,
    RMU_DATA_MSG=0x62,
    RMU_DATA_REQ=0x63,
    MODBUS_DATA_MSG=0x68,
    MODBUS_READ_REQ=0x69,
    MODBUS_READ_RESP=0x6A,
    MODBUS_WRITE_REQ=0x6B,
    MODBUS_WRITE_RESP=0x6C,
    MODBUS_ADDRESS_MSG=0x6E,
    PRODUCT_INFO_MSG=0x6D,
    ACCESSORY_VERSION_REQ=0xEE,
    ECS_DATA_REQ=0x90,
    ECS_DATA_MSG_1=0x55,
    ECS_DATA_MSG_2=0xA0,
    STRING_MSG=0xB1,
    HEATPUMP_REQ=0xF7,
)

Address = Enum(
    Int16ub,
    ECS_S2=0x02,
    # 0x13 = 19, ?
    SMS40=0x16,
    RMU40_S1=0x19,
    RMU40_S2=0x1A,
    RMU40_S3=0x1B,
    RMU40_S4=0x1C,
    MODBUS40=0x20,
    HEATPUMP_1=0x41C9,
    HEATPUMP_2=0x42C9,
    HEATPUMP_3=0x43C9,
    HEATPUMP_4=0x44C9,
    HEATPUMP_5=0x45C9,
    HEATPUMP_6=0x46C9,
    HEATPUMP_7=0x47C9,
    HEATPUMP_8=0x48C9,
)

RmuWriteIndex = Enum(
    Int8ub,
    TEMPORARY_LUX=0x02,
    OPERATIONAL_MODE=0x04,
    FUNCTIONS=0x05,
    TEMPERATURE=0x06,
)

ADDRESS_TO_ROOM_TEMP_COIL = {
    Address.RMU40_S1: 40033,
    Address.RMU40_S2: 40032,
    Address.RMU40_S3: 40031,
    Address.RMU40_S4: 40030,
}


# fmt: off
Response = Struct(
    "start_byte" / Const(0x5C, Int8ub),
    "fields" / RawCopy(
        Struct(
            "address" / Address,
            "cmd" / Command,
            "length" / Int8ub,
            "data" / FixedSized(this.length, Data),
        )
    ),
    "checksum" / Checksum(Int8ub, xor8, this.fields.data),
).compile()


RequestData = Switch(
    this.cmd,
    {
        "ACCESSORY_VERSION_REQ": UnionConstruct(None,
            # Modbus and RMU seem to disagree on how to interpret this
            # data, at least from how it looks in the service info screen
            # on the pump.
            "modbus" / Struct(
                "version" / Int16ul,
                "unknown" / Int8ub,
            ),
            "rmu" / Struct(
                "unknown" / Int8ub,
                "version" / Int16ul,
            ),
        ),
        "RMU_WRITE_REQ": Struct(
            "index" / RmuWriteIndex,
            "value" / Switch(
                lambda this: this.index,
                {
                    "TEMPORARY_LUX": Int8ub,
                    "TEMPERATURE": FixedPoint(Int16ul, 0.1, -0.7),
                    "FUNCTIONS": FlagsEnum(
                        Int8ub,
                        allow_additive_heating=0x01,
                        allow_heating=0x02,
                        allow_cooling=0x04,
                    ),
                    "OPERATIONAL_MODE": Int8ub,
                },
                default=Select(
                    Int16ul,
                    Int8ub,
                )
            )
        ),
        "MODBUS_READ_REQ": Struct(
            "coil_address" / Int16ul,
        ),
        "MODBUS_WRITE_REQ": Struct(
            "coil_address" / Int16ul,
            "value" / Bytes(4),
        )
    },
    default=Bytes(this.length),
)

Request = Struct(
    "fields" / RawCopy(
        Struct(
            "start_byte" / Const(0xC0, Int8ub),
            "cmd" / Command,
            "data" / Prefixed(Int8ub, RequestData)
        )
    ),
    "checksum" / Checksum(Int8ub, xor8, this.fields.data),
).compile()
# fmt: on



from dataclasses import dataclass
import datetime
from typing import Dict, Optional, Union

from nibe.exceptions import NoMappingException, ValidationError

MIN_DATE = datetime.date(2007, 1, 1)
MAX_DATE = MIN_DATE + datetime.timedelta(0xFFFE)


def is_coil_boolean(coil):
    if coil.factor != 1:
        return False

    if coil.min == 0 and coil.max == 1:
        return True

    if coil.mappings and all(k in ["0", "1"] for k in coil.mappings):
        return True

    return False


class Coil:
    """Represents a coil."""

    mappings: Optional[Dict[str, str]] = None
    reverse_mappings: Optional[Dict[str, str]] = None

    def __init__(
        self,
        address: int,
        name: str,
        title: str,
        size: str,
        factor: int = 1,
        write: bool = False,
        **kwargs,
    ):
        r"""Initialize coil.

        :param address: Coil address
        :param name: Coil name
        :param title: Coil title
        :param size: Coil size ("u8", "s8", "u16", "s16", "u32", "s32")
        :param factor: Coil raw value factor (used for value translation)
        :param write: Is coil writable
        :param kwargs:
            - mappings: Coil mappings (used for value translation)
            - info: Coil info
            - unit: Coil measurement unit
            - type: Coil type (number or date)
            - min: Coil raw min value (used for validation)
            - max: Coil raw max value (used for validation)
        """

        assert isinstance(address, int), "Address must be defined"
        assert name, "Name must be defined"
        assert title, "Title must be defined"
        assert factor, "Factor must be defined"

        self.address = address
        self.name = name
        self.title = title
        self.size = size
        self.factor = factor
        self.is_writable = write

        if "mappings" in kwargs:
            mappings = kwargs.pop("mappings")
            assert factor == 1, "When mapping is used factor needs to be 1"
            self.set_mappings(mappings)

        self.info = kwargs.pop("info", None)
        self.unit = kwargs.pop("unit", None)
        self.type = kwargs.pop("type", "number")

        assert self.type in [
            "number",
            "date",
        ], f"Invalid coil type {self.type} for coil {self.name}"

        assert not (
            self.has_mappings and self.type == "date"
        ), f"Date coil {self.name} cannot have mappings"

        self.raw_min = kwargs.pop("min", None)
        self.raw_max = kwargs.pop("max", None)

        self.min = self.raw_min / factor if self.raw_min is not None else None
        self.max = self.raw_max / factor if self.raw_max is not None else None

        self.is_boolean = is_coil_boolean(self)
        if self.is_boolean and not self.has_mappings:
            self.set_mappings({"0": "OFF", "1": "ON"})

        self.is_date = self.type == "date"

        self.other = kwargs

    def set_mappings(self, mappings):
        """Set mappings for value translation."""
        if mappings:
            self.mappings = {k: v.upper() for k, v in mappings.items()}
            self.reverse_mappings = {v.upper(): k for k, v in mappings.items()}
        else:
            self.mappings = None
            self.reverse_mappings = None

    @property
    def has_mappings(self):
        """Return True if mappings are defined."""
        return self.mappings is not None

    def get_mapping_for(self, value: int):
        """Return mapping for value.

        :raises NoMappingException: When no mapping is found"""
        if not self.mappings:
            raise NoMappingException(f"No mappings defined for {self.name}")

        try:
            return self.mappings[str(value)]
        except KeyError:
            raise NoMappingException(
                f"Mapping not found for {self.name} coil for value: {value}"
            )

    def get_reverse_mapping_for(self, value: Union[int, float, str, None]) -> int:
        """Return reverse mapping for value.

        :raises NoMappingException: When no mapping is found"""
        if not isinstance(value, str):
            raise ValidationError(
                f"{self.name} coil value ({value}) is invalid type (str is expected)"
            )

        if not self.reverse_mappings:
            raise NoMappingException(
                f"{self.name} coil has no reverse mappings defined"
            )

        try:
            value = value.upper()
            return int(self.reverse_mappings[str(value)])
        except KeyError:
            raise NoMappingException(
                f"{self.name} coil reverse mapping not found for value: {value}"
            )

    def is_raw_value_valid(self, value: int) -> bool:
        """Return True if provided raw value is valid."""
        if not isinstance(value, int):
            return False

        if self.raw_min is not None and value < self.raw_min:
            return False

        if self.raw_max is not None and value > self.raw_max:
            return False

        return True

    def __repr__(self):
        return f"Coil {self.address}, name: {self.name}, title: {self.title}"


@dataclass
class CoilData:
    """Represents a coil data."""

    coil: Coil
    value: Union[int, float, str, datetime.date, None] = None

    def __repr__(self) -> str:
        return f"Coil {self.coil.name}, value: {self.value}"

    @staticmethod
    def from_mapping(coil: Coil, value: int) -> "CoilData":
        """Create CoilData from raw value using mappings."""
        return CoilData(coil, coil.get_mapping_for(value))

    @staticmethod
    def from_raw_value(coil: Coil, value: Union[int, None]) -> "CoilData":
        """Create CoilData from raw value."""
        if value is None:
            return CoilData(coil, None)
        assert coil.is_raw_value_valid(
            value
        ), f"Raw value {value} is out of range for coil {coil.name}"

        if coil.is_date:
            return CoilData(coil, MIN_DATE + datetime.timedelta(days=value))

        if coil.has_mappings:
            return CoilData.from_mapping(coil, value)

        if coil.factor == 1:
            return CoilData(coil, value)
        else:
            return CoilData(coil, value / coil.factor)

    @property
    def raw_value(self) -> int:
        """Return raw value for coil."""
        if self.coil.is_date:
            assert isinstance(
                self.value, datetime.date
            ), f"Provided value '{self.value}' is invalid type (datetime.date is expected) for {self.coil.name}"

            return (self.value - MIN_DATE).days

        if self.coil.has_mappings:
            return self.coil.get_reverse_mapping_for(self.value)

        assert isinstance(
            self.value, (int, float)
        ), f"Provided value '{self.value}' is invalid type (int or float is supported) for {self.coil.name}"

        raw_value = int(self.value * self.coil.factor)
        assert self.coil.is_raw_value_valid(
            raw_value
        ), f"Value {self.value} is out of range for coil {self.coil.name}"

        return raw_value

    def validate(self) -> None:
        """Validate coil data.

        :raises ValidationError: when validation fails"""
        if self.value is None:
            raise ValidationError(f"Value for {self.coil.name} is not set")

        if self.coil.is_date and not isinstance(self.value, datetime.date):
            raise ValidationError(
                f"{self.coil.name} coil value ({self.value}) is invalid type (expected datetime.date)"
            )

        if self.coil.has_mappings:
            self.coil.get_reverse_mapping_for(
                self.value
            )  # can throw NoMappingException(ValidationException) or AssertionError
            return

        if not isinstance(self.value, (int, float, datetime.date)):
            raise ValidationError(
                f"{self.coil.name} coil value ({self.value}) is invalid type (expected int, float or datetime.date)"
            )

        self._check_value_bounds()

    def _check_value_bounds(self):
        if self.coil.is_date:
            if self.value < MIN_DATE:
                raise ValidationError(
                    f"{self.coil.name} coil value ({self.value}) is smaller than min allowed ({MIN_DATE})"
                )

            if self.value > MAX_DATE:
                raise ValidationError(
                    f"{self.coil.name} coil value ({self.value}) is larger than max allowed ({MAX_DATE})"
                )

        else:
            if self.coil.min is not None and self.value < self.coil.min:
                raise ValidationError(
                    f"{self.coil.name} coil value ({self.value}) is smaller than min allowed ({self.coil.min})"
                )

            if self.coil.max is not None and self.value > self.coil.max:
                raise ValidationError(
                    f"{self.coil.name} coil value ({self.value}) is larger than max allowed ({self.coil.max})"
                )



from __future__ import annotations

from ast import literal_eval
import asyncio
from contextlib import AbstractAsyncContextManager
import io
import logging
from typing import IO

import asyncclick as click
from construct import Const, GreedyRange, Int8ul, RawCopy, Select, Struct, Terminated

from ..coil import CoilData
from ..connection import Connection
from ..connection.modbus import Modbus
from ..connection.nibegw import NibeGW, Request, Response
from ..exceptions import NibeException
from ..heatpump import HeatPump, Model

Ack = Struct("fields" / RawCopy(Struct("Ack" / Const(0x06, Int8ul))))

Nak = Struct("fields" / RawCopy(Struct("Nak" / Const(0x15, Int8ul))))

Block = Select(
    Terminated,
    Response,
    Request,
    Ack,
    Nak,
)

Stream = GreedyRange(Block)


@click.group()
@click.option("-v", "--verbose", count=True)
async def cli(verbose: int):
    if verbose == 0:
        log_level = logging.WARNING
    elif verbose == 1:
        log_level = logging.INFO
    else:
        log_level = logging.DEBUG

    logging.basicConfig(
        format="[%(levelname)-8s] %(message)s",
        level=log_level,
    )
    logging.log(logging.INFO, "Log level set to %r", log_level)


class ConnectionContext(AbstractAsyncContextManager):
    def __init__(self, heatpump: HeatPump, connection: Connection) -> None:
        self.heatpump = heatpump
        self.connection = connection

    async def __aenter__(self) -> ConnectionContext:
        await self.connection.start()
        return self

    async def __aexit__(self, __exc_type, __exc_value, __traceback):
        await self.connection.stop()


@cli.group(chain=True, help="Connect using nibegw protocol")
@click.argument("remote_ip", type=str)
@click.option("--listening_ip", type=str)
@click.option("--listening_port", type=int, default=10090)
@click.option("--remote_read_port", type=int, default=10091)
@click.option("--remote_write_port", type=int, default=10092)
@click.option(
    "--model",
    type=click.Choice([model.name for model in Model.__members__.values()]),
    default=None,
)
@click.pass_context
async def nibegw(
    ctx: click.Context,
    remote_ip: str,
    listening_port: int,
    listening_ip: str | None,
    remote_read_port: int,
    remote_write_port: int,
    model: str | None,
):
    heatpump = HeatPump()
    if model:
        heatpump.model = Model[model]
        await heatpump.initialize()

    connection = NibeGW(
        heatpump=heatpump,
        remote_ip=remote_ip,
        listening_port=listening_port,
        listening_ip=listening_ip,
        remote_read_port=remote_read_port,
        remote_write_port=remote_write_port,
    )

    ctx.obj = await ctx.with_async_resource(ConnectionContext(heatpump, connection))


@cli.group(chain=True, help="Connect using modbus protocol")
@click.argument("remote_ip", type=str)
@click.option("--remote_port", type=int, default=502)
@click.option("--slave_id", type=int, default=1)
@click.option(
    "--model",
    type=click.Choice([model.name for model in Model]),
    default=Model.F1155.name,
)
@click.pass_context
async def modbus(
    ctx: click.Context, remote_ip: str, remote_port: str, model: str, slave_id: int
):
    heatpump = HeatPump(Model[model])
    await heatpump.initialize()
    connection = Modbus(
        heatpump=heatpump,
        url=f"tcp://{remote_ip}:{remote_port}",
        slave_id=slave_id,
    )

    ctx.obj = await ctx.with_async_resource(ConnectionContext(heatpump, connection))


def add_connect_command(command: click.Command):
    nibegw.add_command(command)
    modbus.add_command(command)


@click.command(help="Monitor data sent by pump out of band")
@click.pass_obj
async def monitor(obj: ConnectionContext):
    def on_coil_update(coil_data: CoilData):
        click.echo(coil_data)

    obj.heatpump.subscribe(HeatPump.COIL_UPDATE_EVENT, on_coil_update)

    while True:
        await asyncio.sleep(1)


add_connect_command(monitor)


@click.command()
@click.pass_obj
async def product(obj: ConnectionContext):
    product_info = await obj.connection.read_product_info()
    click.echo(product_info)


add_connect_command(product)


@click.command()
@click.pass_obj
@click.argument("parameter", type=int)
async def read(obj: ConnectionContext, parameter: int, **kwargs):
    coil = obj.heatpump.get_coil_by_address(parameter)
    click.echo(await obj.connection.read_coil(coil))


add_connect_command(read)


@click.command()
@click.pass_obj
@click.argument("parameter", type=int)
@click.argument("value", type=str)
async def write(obj: ConnectionContext, parameter: int, value: str, **kwargs):
    coil = obj.heatpump.get_coil_by_address(parameter)
    if not coil.mappings:
        value = float(value)

    coil_data = CoilData(coil, value)
    await obj.connection.write_coil(coil_data)
    click.echo(coil_data)


add_connect_command(read)


@cli.command()
@click.argument("data", type=str)
@click.option("--type", type=click.Choice(["hex", "bytes"]), default="hex")
async def parse_data(data: str, type: str):
    if type == "hex":
        raw = bytes.fromhex(data)
    elif type == "bytes":
        raw = bytes(literal_eval(data))
    request = Block.parse(raw)
    click.echo(request)


def read_bytes_socat(file: IO):
    lines: list[str] = file.readlines()

    for line in lines:
        if line.startswith("> "):
            continue
        yield from bytes.fromhex(line)


def parse_stream(stream: io.RawIOBase):
    while block := Block.parse_stream(stream):
        yield block


@cli.command()
@click.argument("file", type=click.File())
def parse_file(file: IO):
    with io.BytesIO(bytes(read_bytes_socat(file))) as stream:
        for packet in parse_stream(stream):
            click.echo(packet.fields.value)

        remaining = stream.read()
        if remaining:
            click.echo(f"Remaining: {stream.read()}")


def main():
    try:
        cli()
    except NibeException as exception:
        click.echo(repr(exception))
    except (KeyboardInterrupt, SystemExit):
        pass


if __name__ == "__main__":
    main()



import asyncio
import logging

from async_modbus import modbus_for_url
import async_timeout
from tenacity import retry, retry_if_exception_type, stop_after_attempt
from umodbus.exceptions import ModbusError

from nibe.coil import Coil, CoilData
from nibe.connection import DEFAULT_TIMEOUT, Connection
from nibe.connection.encoders import CoilDataEncoderModbus
from nibe.exceptions import (
    DecodeException,
    ModbusUrlException,
    ReadException,
    ReadIOException,
    ReadTimeoutException,
    ValidationError,
    WriteIOException,
    WriteTimeoutException,
)
from nibe.heatpump import HeatPump

from . import verify_connectivity_read_write_alarm

logger = logging.getLogger("nibe").getChild(__name__)


def split_modbus_data(coil: Coil):
    entity_type = coil.address // 10000
    entity_address = (coil.address % 10000) - 1

    if coil.size in ("s32", "u32"):
        entity_count = 2
    else:
        entity_count = 1

    return entity_type, entity_address, entity_count


class Modbus(Connection):
    """Modbus connection."""

    def __init__(
        self,
        heatpump: HeatPump,
        url,
        slave_id,
        conn_options=None,
        read_retries: int = 3,
        write_retries: int = 3,
    ):
        self._slave_id = slave_id
        self._heatpump = heatpump

        self.read_coil = retry(
            retry=retry_if_exception_type(ReadIOException),
            stop=stop_after_attempt(read_retries),
            reraise=True,
        )(self.read_coil)

        self.write_coil = retry(
            retry=retry_if_exception_type(WriteIOException),
            stop=stop_after_attempt(write_retries),
            reraise=True,
        )(self.write_coil)

        try:
            self._client = modbus_for_url(url, conn_options)
        except ValueError as exc:
            raise ModbusUrlException(str(exc)) from exc

        self.coil_encoder = CoilDataEncoderModbus(heatpump.word_swap)

    async def stop(self) -> None:
        await self._client.stream.close()

    async def read_coil(self, coil: Coil, timeout: float = DEFAULT_TIMEOUT) -> CoilData:
        logger.debug("Sending read request")
        entity_type, entity_number, entity_count = split_modbus_data(coil)

        try:
            async with async_timeout.timeout(timeout):
                if entity_type == 3:
                    result = await self._client.read_input_registers(
                        slave_id=self._slave_id,
                        starting_address=entity_number,
                        quantity=entity_count,
                    )
                elif entity_type == 4:
                    result = await self._client.read_holding_registers(
                        slave_id=self._slave_id,
                        starting_address=entity_number,
                        quantity=entity_count,
                    )
                elif entity_type == 1:
                    result = await self._client.read_discrete_inputs(
                        slave_id=self._slave_id,
                        starting_address=entity_number,
                        quantity=entity_count,
                    )
                elif entity_type == 0:
                    result = await self._client.read_coils(
                        slave_id=self._slave_id,
                        starting_address=entity_number,
                        quantity=entity_count,
                    )
                else:
                    raise ReadException(f"Unsupported entity type {entity_type}")

            coil_data = self.coil_encoder.decode(coil, result)

            logger.info(coil_data)
            self._heatpump.notify_coil_update(coil_data)
        except ModbusError as exc:
            raise ReadIOException(
                f"Error '{str(exc)}' reading {coil.name} starting: {entity_number} count: {entity_count} from: {self._slave_id}"
            ) from exc
        except asyncio.TimeoutError as exc:
            raise ReadTimeoutException(
                f"Timeout waiting for read response for {coil.name}"
            ) from exc
        except (DecodeException, ValidationError) as exc:
            raise ReadException(f"Failed decoding response for {coil.name}") from exc

        return coil_data

    async def write_coil(
        self, coil_data: CoilData, timeout: float = DEFAULT_TIMEOUT
    ) -> None:
        coil = coil_data.coil
        assert coil.is_writable, f"{coil.name} is not writable"

        entity_type, entity_number, entity_count = split_modbus_data(coil)
        try:
            values = self.coil_encoder.encode(coil_data)

            logger.debug("Sending write request")
            async with async_timeout.timeout(timeout):
                if entity_type == 4:
                    result = await self._client.write_registers(
                        slave_id=self._slave_id,
                        starting_address=entity_number,
                        values=values,
                    )
                elif entity_type == 0:
                    result = await self._client.write_coils(
                        slave_id=self._slave_id,
                        starting_address=entity_number,
                        values=values,
                    )
                else:
                    raise ReadIOException(f"Unsupported entity type {entity_type}")

            if not result:
                raise WriteIOException(f"Heatpump denied writing {coil.name}")
            else:
                logger.info(f"Write succeeded for {coil.name}")
        except ValidationError as exc:
            raise WriteIOException(
                f"Error validating {coil.name} coil value: {str(exc)}"
            ) from exc
        except ModbusError as exc:
            raise WriteIOException(
                f"Error '{str(exc)}' writing {coil.name} starting: {entity_number} count: {entity_count} to: {self._slave_id}"
            ) from exc
        except asyncio.TimeoutError:
            raise WriteTimeoutException(
                f"Timeout waiting for write feedback for {coil.name}"
            )

    async def verify_connectivity(self):
        await verify_connectivity_read_write_alarm(self, self._heatpump)



from collections import defaultdict
import logging
from typing import Callable

logger = logging.getLogger("nibe").getChild(__name__)


class EventServer:
    _listeners: defaultdict[str, list[Callable[..., None]]]

    def __init__(self):
        self._listeners = defaultdict(list)

    def notify_event_listeners(self, event_name: str, *args, **kwargs):
        """Notify all listeners of an event"""
        for listener in self._listeners[event_name]:
            try:
                listener(*args, **kwargs)
            except Exception as e:
                logger.exception(e)

    def subscribe(self, event_name: str, callback: Callable[..., None]):
        """Subscribe to an event"""
        self._listeners[event_name].append(callback)



from abc import abstractmethod
from binascii import hexlify
from typing import Generic, List, Optional, TypeVar, Union

from construct import (
    Construct,
    ConstructError,
    Int8sl,
    Int8ul,
    Int16sl,
    Int16ul,
    Int32sl,
    Int32ul,
    Padded,
)

from nibe.coil import Coil, CoilData
from nibe.exceptions import DecodeException, EncodeException, ValidationError
from nibe.parsers import WordSwapped

parser_map = {
    "u8": Int8ul,
    "s8": Int8sl,
    "u16": Int16ul,
    "s16": Int16sl,
    "u32": Int32ul,
    "s32": Int32sl,
}

parser_map_word_swapped = parser_map.copy()
parser_map_word_swapped.update(
    {
        "u32": WordSwapped(Int32ul),
        "s32": WordSwapped(Int32sl),
    }
)

integer_limit = {
    "u8": 0xFF,
    "s8": -0x80,
    "u16": 0xFFFF,
    "s16": -0x8000,
    "u32": 0xFFFFFFFF,
    "s32": -0x80000000,
}


_RawDataT = TypeVar("_RawDataT")


class CoilDataEncoder(Generic[_RawDataT]):
    """Encode and decode coil data."""

    word_swap: Optional[bool] = None

    def __init__(self, word_swap: Optional[bool] = None):
        self.word_swap = word_swap

    @abstractmethod
    def encode_raw_value(self, size: str, raw_value: int) -> _RawDataT:
        pass

    def encode(self, coil_data: CoilData) -> _RawDataT:
        """Encode coil data to bytes.

        :raises EncodeException: If encoding fails"""
        try:
            coil_data.validate()
            return self.encode_raw_value(coil_data.coil.size, coil_data.raw_value)
        except (ValueError, ConstructError, ValidationError) as e:
            raise EncodeException(
                f"Failed to encode {coil_data.coil.name} coil for value: {coil_data.value}, exception: {e}"
            )

    @abstractmethod
    def decode_raw_value(self, size: str, raw: _RawDataT) -> int:
        pass

    def decode(self, coil: Coil, raw: _RawDataT) -> CoilData:
        """Decode coil data from bytes.

        :raises DecodeException: If decoding fails"""
        try:
            value = self.decode_raw_value(coil.size, raw)
            if self._is_hitting_integer_limit(coil.size, value):
                value = None

            return CoilData.from_raw_value(coil, value)

        except (ValueError, AssertionError, ConstructError, ValidationError) as e:
            raise DecodeException(
                f"Failed to decode {coil.name} coil from raw: {hexlify(raw).decode('utf-8')}, exception: {e}"
            ) from e

    def _is_hitting_integer_limit(self, size: str, int_value: int):
        limit = integer_limit[size]
        if limit < 0:
            return int_value <= limit
        return int_value >= limit


class CoilDataEncoderNibeGw(CoilDataEncoder[bytes]):
    """Encode and decode coil data."""

    word_swap: Optional[bool] = None

    def __init__(self, word_swap: Optional[bool] = None):
        self.word_swap = word_swap

    def encode_raw_value(self, size: str, raw_value: int) -> bytes:
        """Encode coil data to bytes."""
        return self._pad(self._get_parser(size), raw_value)

    def decode_raw_value(self, size: str, raw: bytes) -> int:
        """Decode coil data from bytes."""
        parser = self._get_parser(size)
        assert parser.sizeof() <= len(
            raw
        ), f"Invalid raw data size: given {len(raw)}, expected at least {parser.sizeof()}"
        value = parser.parse(raw)
        return value

    def _get_parser(self, size: str) -> Construct:
        if size in ["u32", "s32"] and self.word_swap is None:
            raise ValueError("Word swap is not set, cannot parse 32 bit integers")

        if self.word_swap:  # yes, it is visa versa
            return parser_map[size]
        else:
            return parser_map_word_swapped[size]

    def _pad(self, parser: Construct, value: int) -> bytes:
        return Padded(4, parser).build(value)


class CoilDataEncoderModbus(CoilDataEncoder[List[int]]):
    word_swap: Optional[bool] = None

    def __init__(self, word_swap: Optional[bool] = None):
        self.word_swap = word_swap

    def encode_raw_value(self, size: str, raw_value: int) -> List[int]:
        signed = size in ("s32", "s16", "s8")

        raw_bytes = raw_value.to_bytes(8, "little", signed=signed)
        if size in ("s32", "u32"):
            if self.word_swap:
                return [
                    int.from_bytes(raw_bytes[0:2], "little", signed=False),
                    int.from_bytes(raw_bytes[2:4], "little", signed=False),
                ]
            else:
                return [
                    int.from_bytes(raw_bytes[2:4], "little", signed=False),
                    int.from_bytes(raw_bytes[0:2], "little", signed=False),
                ]
        elif size in ("s16", "u16", "s8", "u8"):
            return [int.from_bytes(raw_bytes[0:2], "little", signed=False)]
        raise ValueError("Unknown coil encoding")

    def decode_raw_value(self, size: str, raw: List[int]) -> Union[int, None]:
        signed = size in ("s32", "s16", "s8")

        if not self.word_swap:
            raw = reversed(raw)
        raw_bytes = [byte for value in raw for byte in value.to_bytes(2, "little")]
        raw_value = int.from_bytes(raw_bytes, byteorder="little", signed=signed)

        if self._is_hitting_integer_limit(size, raw_value):
            return None

        return raw_value
